#encoding: utf-8
'''
Various utilities for using the ISLE dictionary with praat textgrids

see
**examples/alignment_example.py**
**examples/syllabify_textgrid.py**
'''


class OptionalFeatureError(ImportError):
    
    def __str__(self):
        return "ERROR: You must have praatio installed to use pysle.praatTools"

try:
    from praatio import tgio
    from praatio import praatio_scripts
except ImportError:
    raise OptionalFeatureError()

from pysle import isletool
from pysle import pronunciationtools


def spellCheckTextgrid(tg, targetTierName, newTierName, isleDict,
                       printEntries=False):
    '''
    Spell check words by using the praatio spellcheck function
    
    Incorrect items are noted in a new tier and optionally
        printed to the screen
    '''
    
    def checkFunc(word):
        try:
            isleDict.lookup(word)
        except isletool.WordNotInISLE:
            returnVal = False
        else:
            returnVal = True
        
        return returnVal
    
    tg = praatio_scripts.spellCheckEntries(tg, targetTierName, newTierName,
                                           checkFunc, printEntries)
    
    return tg
    

def naiveWordAlignment(tg, utteranceTierName, wordTierName, isleDict,
                       phoneHelperTierName=None,
                       removeOverlappingSegments=False):
    '''
    Performs naive alignment for utterances in a textgrid
    
    Naive alignment gives each segment equal duration.  Word duration is
    determined by the duration of an utterance and the number of phones in
    the word.
    
    By 'utterance' I mean a string of words separated by a space bounded
    in time eg (0.5, 1.5, "he said he likes ketchup").
    
    phoneHelperTierName - creates a tier that is parallel to the word tier.
                          However, the labels are the phones for the word,
                          rather than the word
    removeOverlappingSegments - remove any labeled words or phones that
                                fall under labeled utterances
    '''
    utteranceTier = tg.tierDict[utteranceTierName]
    
    wordTier = None
    if wordTierName in tg.tierNameList:
        wordTier = tg.tierDict[wordTierName]
    
    # Load in the word tier, if it exists:
    wordEntryList = []
    phoneEntryList = []
    if wordTier is not None:
        if removeOverlappingSegments:
            for startT, stopT, _ in utteranceTier.entryList:
                wordTier = wordTier.eraseRegion(startT, stopT,
                                                'truncate', False)
        wordEntryList = wordTier.entryList

    # Do the naive alignment
    for startT, stopT, label in utteranceTier.entryList:
        wordList = label.split()

        # Get the list of phones in each word
        superPhoneList = []
        numPhones = 0
        i = 0
        while i < len(wordList):
            word = wordList[i]
            try:
                firstSyllableList = isleDict.lookup(word)[0][0][0]
            except isletool.WordNotInISLE:
                wordList.pop(i)
                continue
            phoneList = [phone for syllable in firstSyllableList
                         for phone in syllable]
            superPhoneList.append(phoneList)
            numPhones += len(phoneList)
            i += 1
        
        # Get the naive alignment for words, if alignment doesn't
        # already exist for words
        subWordEntryList = []
        subPhoneEntryList = []
        if wordTier is not None:
            subWordEntryList = wordTier.crop(startT, stopT,
                                             "truncated", False).entryList
        
        if len(subWordEntryList) == 0:
            wordStartT = startT
            phoneDur = (stopT - startT) / float(numPhones)
            for i, word in enumerate(wordList):
                phoneListTxt = " ".join(superPhoneList[i])
                wordStartT = wordStartT
                wordEndT = wordStartT + (phoneDur * len(superPhoneList[i]))
                subWordEntryList.append((wordStartT, wordEndT, word))
                subPhoneEntryList.append((wordStartT, wordEndT, phoneListTxt))
                wordStartT = wordEndT
        
        wordEntryList.extend(subWordEntryList)
        phoneEntryList.extend(subPhoneEntryList)
    
    # Replace or add the word tier
    newWordTier = tgio.IntervalTier(wordTierName,
                                    wordEntryList,
                                    tg.minTimestamp,
                                    tg.maxTimestamp)
    if wordTier is not None:
        tg.replaceTier(wordTierName, newWordTier)
    else:

        tg.addTier(newWordTier)
        
    # Add the phone tier
    # This is mainly used as an annotation tier
    if phoneHelperTierName is not None and len(phoneEntryList) > 0:
        newPhoneTier = tgio.IntervalTier(phoneHelperTierName,
                                         phoneEntryList,
                                         tg.minTimestamp,
                                         tg.minTimestamp)
        if phoneHelperTierName in tg.tierNameList:
            tg.replaceTier(phoneHelperTierName, newPhoneTier)
        else:
            tg.addTier(newPhoneTier)
    
    return tg


def naivePhoneAlignment(tg, wordTierName, phoneTierName, isleDict,
                        removeOverlappingSegments=False):
    '''
    Performs naive alignment for words in a textgrid
    
    Naive alignment gives each segment equal duration.
    Phone duration is determined by the duration of the word
    and the number of phones.
    
    removeOverlappingSegments - remove any labeled words or phones that
                                fall under labeled utterances
    '''
    wordTier = tg.tierDict[wordTierName]
    
    phoneTier = None
    if phoneTierName in tg.tierNameList:
        phoneTier = tg.tierDict[phoneTierName]
    
    # Load in the phone tier, if it exists:
    phoneEntryList = []
    if phoneTier is not None:
        if removeOverlappingSegments:
            for startT, stopT, _ in wordTier.entryList:
                phoneTier = phoneTier.eraseRegion(startT, stopT,
                                                  'truncate', False)
        phoneEntryList = phoneTier.entryList

    # Do the naive alignment
    for wordStartT, wordEndT, word in wordTier.entryList:
        
        # Get the list of phones in this word
        try:
            firstSyllableList = isleDict.lookup(word)[0][0][0]
        except isletool.WordNotInISLE:
            continue

        phoneList = [phone for syllable in firstSyllableList
                     for phone in syllable]
        for char in [u'ˈ', u'ˌ']:
            phoneList = [phone.replace(char, '') for phone in phoneList]
        
        # Get the naive alignment for phones, if alignment doesn't
        # already exist for phones
        subPhoneEntryList = []
        if phoneTier is not None:
            subPhoneEntryList = phoneTier.crop(wordStartT, wordEndT,
                                               "truncated", False).entryList
        
        if len(subPhoneEntryList) == 0:
            phoneDur = (wordEndT - wordStartT) / len(phoneList)
            
            phoneStartT = wordStartT
            for phone in phoneList:
                phoneEndT = phoneStartT + phoneDur
                subPhoneEntryList.append((phoneStartT, phoneEndT, phone))
                phoneStartT = phoneEndT

        phoneEntryList.extend(subPhoneEntryList)
    
    # Replace or add the phone tier
    newPhoneTier = tgio.IntervalTier(phoneTierName,
                                     phoneEntryList,
                                     tg.minTimestamp,
                                     tg.maxTimestamp)
    if phoneTier is not None:
        tg.replaceTier(phoneTierName, newPhoneTier)
    else:

        tg.addTier(newPhoneTier)
    
    return tg


def syllabifyTextgrid(isleDict, tg, wordTierName, phoneTierName,
                      skipLabelList=None, startT=None, stopT=None):
    '''
    Given a textgrid, syllabifies the phones in the textgrid
    
    skipLabelList allows you to skip labels without generating warnings
    (e.g. '', 'sp', etc.)
    
    The textgrid must have a word tier and a phone tier.
    
    Returns a textgrid with only two tiers containing syllable information
    (syllabification of the phone tier and a tier marking word-stress).
    '''
    minT = tg.minTimestamp
    maxT = tg.maxTimestamp
    
    wordTier = tg.tierDict[wordTierName]
    phoneTier = tg.tierDict[phoneTierName]
    
    if skipLabelList is None:
        skipLabelList = []
    
    syllableEntryList = []
    tonicSEntryList = []
    tonicPEntryList = []
    
    if startT is not None or stopT is not None:
        if startT is None:
            startT = minT
        if stopT is None:
            stopT = maxT
    
        wordTier = wordTier.crop(startT, stopT, "truncated", False)
    
    for start, stop, word in wordTier.entryList:
        
        if word in skipLabelList:
            continue
        
        subPhoneTier = phoneTier.crop(start, stop, "strict", False)
        
        # entry = (start, stop, phone)
        phoneList = [entry[2] for entry in subPhoneTier.entryList
                     if entry[2] != '']
        phoneList = phoneList
        
        try:
            sylTmp = pronunciationtools.findBestSyllabification(isleDict,
                                                                word,
                                                                phoneList)
        except isletool.WordNotInISLE:
            print("Word ('%s') not is isle -- skipping syllabification" % word)
            continue
        except (pronunciationtools.NullPronunciationError):
            print("Word ('%s') has no provided pronunciation" % word)
            continue
        except AssertionError:
            print("Unable to syllabify '%s'" % word)
            continue
        

        stressI = sylTmp[0]
        stressJ = sylTmp[1]
        syllableList = sylTmp[2]
            
        stressedPhone = None
        if stressI is not None and stressJ is not None:
            stressedPhone = syllableList[stressI][stressJ]
            syllableList[stressI][stressJ] += u"ˈ"

        i = 0
        for k, syllable in enumerate(syllableList):
            
            # Create the syllable tier entry
            j = len(syllable)
            stubEntryList = subPhoneTier.entryList[i:i + j]
            i += j
            
            # The whole syllable was deleted
            if len(stubEntryList) == 0:
                continue
            
            syllableStart = stubEntryList[0][0]
            syllableEnd = stubEntryList[-1][1]
            label = "-".join([entry[2] for entry in stubEntryList])
        
            syllableEntryList.append((syllableStart, syllableEnd, label))
            
            # Create the tonic syllable tier entry
            if k == stressI:
                tonicSEntryList.append((syllableStart, syllableEnd, 'T'))
            
            # Create the tonic phone tier entry
            if k == stressI:
                syllablePhoneTier = phoneTier.crop(syllableStart,
                                                   syllableEnd,
                                                   "strict", False)
            
                phoneList = [entry for entry in syllablePhoneTier.entryList
                             if entry[2] != '']
                justPhones = [phone for _, _, phone in phoneList]
                cvList = pronunciationtools._prepPronunciation(justPhones)
                
                try:
                    tmpStressJ = cvList.index('V')
                except ValueError:
                    for char in [u'r', u'n', u'l']:
                        if char in cvList:
                            tmpStressJ = cvList.index(char)
                            break
                        
                phoneStart, phoneEnd = phoneList[tmpStressJ][:2]
                tonicPEntryList.append((phoneStart, phoneEnd, 'T'))
    
    # Create a textgrid with the two syllable-level tiers
    syllableTier = tgio.IntervalTier('syllable', syllableEntryList,
                                     minT, maxT)
    tonicSTier = tgio.IntervalTier('tonicSyllable', tonicSEntryList,
                                   minT, maxT)
    tonicPTier = tgio.IntervalTier('tonicVowel', tonicPEntryList,
                                   minT, maxT)
    
    syllableTG = tgio.Textgrid()
    syllableTG.addTier(syllableTier)
    syllableTG.addTier(tonicSTier)
    syllableTG.addTier(tonicPTier)

    return syllableTG



#encoding: utf-8
'''
The main interface for working with the ISLE dictionary.

Can be used to run searches based on written form or
pronunciation ('cat' vs [kat]).

Also has various utility functions.

see
**examples/basic_examples.py**
**examples/dictionary_search.py**
'''

import io
import re


charList = [u'#', u'.', u'aʊ', u'b', u'd', u'dʒ', u'ei', u'f', u'g',
            u'h', u'i', u'j', u'k', u'l', u'm', u'n', u'oʊ', u'p',
            u'r', u's', u't', u'tʃ', u'u', u'v', u'w', u'z', u'æ',
            u'ð', u'ŋ', u'ɑ', u'ɑɪ', u'ɔ', u'ɔi', u'ə', u'ɚ', u'ɛ', u'ɝ',
            u'ɪ', u'ɵ', u'ɹ', u'ʃ', u'ʊ', u'ʒ', u'æ', u'ʌ', ]

diacriticList = [u'˺', u'ˌ', u'̩', u'̃', u'ˈ', ]

monophthongList = [u'u', u'æ', u'ɑ', u'ɔ', u'ə', u'i', u'ɛ', u'ɪ', u'ʊ', u'ʌ', u'a', u'e', u'o', ]

diphthongList = [u'ɑɪ', u'aʊ', u'ei', u'ɔi', u'oʊ', u'ae']

syllabicConsonantList = [u'l̩', u'n̩', u'ɚ', u'ɝ']

# ISLE words are part of speech tagged using the Penn Part of Speech Tagset
posList = ['cc', 'cd', 'dt', 'fw', 'in', 'jj', 'jjr', 'jjs', 'ls', 'md',
           'nn', 'nnd', 'nnp', 'nnps', 'nns', 'pdt', 'prp', 'punc', 'rb',
           'rbr', 'rbs', 'rp', 'sym', 'to', 'uh', 'vb', 'vbd', 'vbg', 'vbn',
           'vbp', 'vbz', 'vpb', 'wdt', 'wp', 'wrb']

vowelList = monophthongList + diphthongList + syllabicConsonantList


def isVowel(char):
    return any([vowel in char for vowel in vowelList])


def sequenceMatch(matchChar, searchStr):
    return matchChar in searchStr


class WordNotInISLE(Exception):
    
    def __init__(self, word):
        super(WordNotInISLE, self).__init__()
        self.word = word
        
    def __str__(self):
        return ("Word '%s' not in ISLE dictionary.  "
                "Please add it to continue." % self.word)


class LexicalTool():
    
    def __init__(self, islePath):
        '''
        
        
        self.data: the pronunciation data {(word, pronunciationList),}
        self.dataExtra: pos and other info {(word, infoList),}
        '''
        self.islePath = islePath
        self.data = self._buildDict()
    
    def _buildDict(self):
        '''
        Builds the isle textfile into a dictionary for fast searching
        '''
        lexDict = {}
        with io.open(self.islePath, "r", encoding='utf-8') as fd:
            wordList = [line.rstrip('\n') for line in fd]
        
        for row in wordList:
            word, pronunciation = row.split(" ", 1)
            word, extraInfo = word.split("(", 1)
            
            extraInfo = extraInfo.replace(")", "")
            extraInfoList = [segment for segment in extraInfo.split(",")
                             if ("_" not in segment and "+" not in segment and
                                 ':' not in segment and segment != '')]
            
            lexDict.setdefault(word, [])
            lexDict[word].append((pronunciation, extraInfoList))
        
        return lexDict
    
    def lookup(self, word):
        '''
        Lookup a word and receive a list of syllables and stressInfo
        
        Output example for the word 'another' which has two pronunciations
        [(([[u'ə'], [u'n', u'ˈʌ'], [u'ð', u'ɚ']], [1], [1]),
          ([[u'ə'], [u'n', u'ˈʌ'], [u'ð', u'ə', u'ɹ']], [1], [1]))]
        '''
        
        # All words must be lowercase with no extraneous whitespace
        word = word.lower()
        word = word.strip()
        
        pronList = self.data.get(word, None)
        
        if pronList is None:
            raise WordNotInISLE(word)
        else:
            pronList = [_parsePronunciation(pronunciationStr)
                        for pronunciationStr, _ in pronList]
            pronList = list(zip(*pronList))
        
        return pronList

    def search(self, matchStr, numSyllables=None, wordInitial='ok',
               wordFinal='ok', spanSyllable='ok', stressedSyllable='ok',
               multiword='ok', pos=None):
        '''
        for help on isletool.LexicalTool.search(), see see isletool.search()
        '''
        return search(self.data.items(), matchStr, numSyllables=numSyllables,
                      wordInitial=wordInitial, wordFinal=wordFinal,
                      spanSyllable=spanSyllable,
                      stressedSyllable=stressedSyllable,
                      multiword=multiword, pos=pos)


def _prepRESearchStr(matchStr, wordInitial='ok', wordFinal='ok',
                     spanSyllable='ok', stressedSyllable='ok'):
    '''
    Prepares a user's RE string for a search
    '''
    
    # Protect sounds that are two characters
    # After this we can assume that each character represents a sound
    # (We'll revert back when we're done processing the RE)
    replList = [(u'ei', u'9'), (u'tʃ', u'='), (u'oʊ', u'~'),
                (u'dʒ', u'@'), (u'aʊ', u'%'), (u'ɑɪ', u'&'),
                (u'ɔi', u'$')]

    # Add to the replList
    currentReplNum = 0
    startI = 0
    for left, right in (('(', ')'), ('[', ']')):
        while True:
            try:
                i = matchStr.index(left, startI)
            except ValueError:
                break
            j = matchStr.index(right, i) + 1
            replList.append((matchStr[i:j], str(currentReplNum)))
            currentReplNum += 1
            startI = j
    
    for charA, charB in replList:
        matchStr = matchStr.replace(charA, charB)
    
    # Characters to check between all other characters
    # Don't check between all other characters if the character is already
    # in the search string or
    interleaveStr = None
    stressOpt = (stressedSyllable == 'ok' or stressedSyllable == 'only')
    spanOpt = (spanSyllable == 'ok' or spanSyllable == 'only')
    if stressOpt and spanOpt:
        interleaveStr = u"\\.?ˈ?"
    elif stressOpt:
        interleaveStr = u"ˈ?"
    elif spanOpt:
        interleaveStr = u"\\.?"
    
    if interleaveStr is not None:
        matchStr = interleaveStr.join(matchStr)
    
    # Setting search boundaries
    # We search on '[^\.#]' and not '.' so that the search doesn't span
    # multiple syllables or words
    if wordInitial == 'only':
        matchStr = u'#' + matchStr
    elif wordInitial == 'no':
        # Match the closest preceeding syllable.  If there is none, look
        # for word boundary plus at least one other character
        matchStr = u'(?:\\.[^\\.#]*?|#[^\\.#]+?)' + matchStr
    else:
        matchStr = u'[#\\.][^\\.#]*?' + matchStr
    
    if wordFinal == 'only':
        matchStr = matchStr + u'#'
    elif wordFinal == 'no':
        matchStr = matchStr + u"(?:[^\\.#]*?\\.|[^\\.#]+?#)"
    else:
        matchStr = matchStr + u'[^\\.#]*?[#\\.]'
    
    # For sounds that are designated two characters, prevent
    # detecting those sounds if the user wanted a sound
    # designated by one of the contained characters
    
    # Forward search ('a' and not 'ab')
    insertList = []
    for charA, charB in [(u'e', u'i'), (u't', u'ʃ'), (u'd', u'ʒ'),
                         (u'o', u'ʊ'), (u'a', u'ʊ|ɪ'), (u'ɔ', u'i'), ]:
        startI = 0
        while True:
            try:
                i = matchStr.index(charA, startI)
            except ValueError:
                break
            if matchStr[i + 1] != charB:
                forwardStr = u'(?!%s)' % charB
#                 matchStr = matchStr[:i + 1] + forwardStr + matchStr[i + 1:]
                startI = i + 1 + len(forwardStr)
                insertList.append((i + 1, forwardStr))
        
    # Backward search ('b' and not 'ab')
    for charA, charB in [(u't', u'ʃ'), (u'd', u'ʒ'),
                         (u'a|o', u'ʊ'), (u'e|ɔ', u'i'), (u'ɑ' u'ɪ'), ]:
        startI = 0
        while True:
            try:
                i = matchStr.index(charB, startI)
            except ValueError:
                break
            if matchStr[i - 1] != charA:
                backStr = u'(?<!%s)' % charA
#                 matchStr = matchStr[:i] + backStr + matchStr[i:]
                startI = i + 1 + len(backStr)
                insertList.append((i, backStr))
                
    insertList.sort()
    for i, insertStr in insertList[::-1]:
        matchStr = matchStr[:i] + insertStr + matchStr[i:]
    
    # Revert the special sounds back from 1 character to 2 characters
    for charA, charB in replList:
        matchStr = matchStr.replace(charB, charA)

    # Replace special characters
    replDict = {"D": u"(?:t(?!ʃ)|d(?!ʒ)|[sz])",  # dentals
                "F": u"[ʃʒfvszɵðh]",  # fricatives
                "S": u"(?:t(?!ʃ)|d(?!ʒ)|[pbkg])",  # stops
                "N": u"[nmŋ]",  # nasals
                "R": u"[rɝɚ]",  # rhotics
                "V": u"(?:aʊ|ei|oʊ|ɑɪ|ɔi|[iuæɑɔəɛɪʊʌ]):?",  # vowels
                "B": u"\\.",  # syllable boundary
                }

    for char, replStr in replDict.items():
        matchStr = matchStr.replace(char, replStr)

    return matchStr


def search(searchList, matchStr, numSyllables=None, wordInitial='ok',
           wordFinal='ok', spanSyllable='ok', stressedSyllable='ok',
           multiword='ok', pos=None):
    '''
    Searches for words in searchList that match the specified pronunciation; uses regular expressions
    
    wordInitial, wordFinal, spanSyllable, stressSyllable, and multiword
    can take three different values: 'ok', 'only', or 'no'.
    
    pos: a tag in the Penn Part of Speech tagset
        see isletool.posList for the full list of possible tags
    
    Special search characters:
    'D' - any dental; 'F' - any fricative; 'S' - any stop
    'V' - any vowel; 'N' - any nasal; 'R' - any rhotic
    '#' - word boundary
    'B' - syllable boundary
    '.' - anything
    
    For advanced queries:
    Regular expression syntax applies, so if you wanted to search for any
    word ending with a vowel or rhotic, matchStr = '(?:VR)#', '[VR]#', etc.

    Compared with LexicalTool().search(), this function can be used to search through a smaller
    set of data than the entire ISLEdict dictionary.
    '''
    # Run search for words
    
    matchStr = _prepRESearchStr(matchStr, wordInitial, wordFinal,
                                spanSyllable, stressedSyllable)
    
    compiledRE = re.compile(matchStr)
    retList = []
    for word, pronList in searchList:
        newPronList = []
        for pron, posList in pronList:
            searchPron = pron.replace(",", "").replace(" ", "")
            
            # Search for pos
            if pos is not None:
                if pos not in posList:
                    continue
            
            # Ignore diacritics for now:
            for diacritic in diacriticList:
                if diacritic not in matchStr:
                    searchPron = searchPron.replace(diacritic, "")
            
            if numSyllables is not None:
                if numSyllables != searchPron.count('.') + 1:
                    continue
            
            # Is this a compound word?
            if multiword == 'only':
                if searchPron.count('#') == 2:
                    continue
            elif multiword == 'no':
                if searchPron.count('#') > 2:
                    continue
            
            matchList = compiledRE.findall(searchPron)
            if len(matchList) > 0:
                if stressedSyllable == 'only':
                    if all([u"ˈ" not in match for match in matchList]):
                        continue
                if stressedSyllable == 'no':
                    if all([u"ˈ" in match for match in matchList]):
                        continue
                
                # For syllable spanning, we check if there is a syllable
                # marker inside (not at the border) of the match.
                if spanSyllable == 'only':
                    if all(["." not in txt[1:-1] for txt in matchList]):
                        continue
                if spanSyllable == 'no':
                    if all(["." in txt[1:-1] for txt in matchList]):
                        continue
                newPronList.append((pron, posList))
        
        if len(newPronList) > 0:
            retList.append((word, newPronList))
    
    retList.sort()
    return retList


def _parsePronunciation(pronunciationStr):
    '''
    Parses the pronunciation string
    
    Returns the list of syllables and a list of primary and
    secondary stress locations
    '''
    retList = []
    for syllableTxt in pronunciationStr.split("#"):
        if syllableTxt == "":
            continue
        syllableList = [x.split() for x in syllableTxt.split(' . ')]
        
        # Find stress
        stressedSyllableList = []
        stressedPhoneList = []
        for i, syllable in enumerate(syllableList):
            for j, phone in enumerate(syllable):
                if u"ˈ" in phone:
                    stressedSyllableList.insert(0, i)
                    stressedPhoneList.insert(0, j)
                    break
                elif u'ˌ' in phone:
                    stressedSyllableList.append(i)
                    stressedPhoneList.append(j)
        
        retList.append((syllableList, stressedSyllableList, stressedPhoneList))
    
    return retList
            
            
def getNumPhones(isleDict, word, maxFlag):
    '''
    Get the number of syllables and phones in this word
    
    If maxFlag=True, use the longest pronunciation.  Otherwise, take the
    average length.
    '''
    phoneCount = 0
    syllableCount = 0
    
    syllableCountList = []
    phoneCountList = []
    
    wordList = isleDict.lookup(word)
    entryList = zip(*wordList)
    
    for lookupResultList in entryList:
        syllableList = []
        for wordSyllableList in lookupResultList:
            syllableList.extend(wordSyllableList)
        
        syllableCountList.append(len(syllableList))
        phoneCountList.append(len([phon for phoneList in syllableList for
                                   phon in phoneList]))
    
    # The average number of phones for all possible pronunciations
    #    of this word
    if maxFlag is True:
        syllableCount += max(syllableCountList)
        phoneCount += max(phoneCountList)
    else:
        syllableCount += (sum(syllableCountList) /
                          float(len(syllableCountList)))
        phoneCount += sum(phoneCountList) / float(len(phoneCountList))
    
    return syllableCount, phoneCount


def findOODWords(isleDict, wordList):
    '''
    Returns all of the out-of-dictionary words found in a list of utterances
    '''
    oodList = []
    for word in wordList:
        try:
            isleDict.lookup(word)
        except WordNotInISLE:
            oodList.append(word)
                
    oodList = list(set(oodList))
    oodList.sort()
    
    return oodList


def autopair(isleDict, wordList):
    '''
    Tests whether adjacent words are OOD or not
    
    It returns complete wordLists with the matching words replaced.
    Each match yields one sentence.
    
    e.g.
    red ball chaser
    would return
    [[red_ball chaser], [red ball_chaser]], [0, 1]
    
    if 'red_ball' and 'ball_chaser' were both in the dictionary
    '''
    
    newWordList = [("%s_%s" % (wordList[i], wordList[i + 1]), i)
                   for i in range(0, len(wordList) - 1)]
    
    sentenceList = []
    indexList = []
    for word, i in newWordList:
        if word in isleDict.data:
            sentenceList.append(wordList[:i] + [word, ] + wordList[i + 1:])
            indexList.append(i)
    
    return sentenceList, indexList



#encoding: utf-8
'''
Code for comparing and aligning pronunciation data with pronunciations found in the ISLE dictionary.

see **examples/pronunciationtools_examples.py**
'''

import itertools
import copy

from pysle import isletool


class TooManyVowelsInSyllable(Exception):
    
    def __init__(self, syllable, syllableCVMapped):
        super(TooManyVowelsInSyllable, self).__init__()
        self.syllable = syllable
        self.syllableCVMapped = syllableCVMapped
        
    def __str__(self):
        errStr = ("Error: syllable '%s' found to have more than "
                  "one vowel.\n This was the CV mapping: '%s'")
        syllableStr = u"".join(self.syllable)
        syllableCVStr = u"".join(self.syllableCVMapped)
        
        return errStr % (syllableStr, syllableCVStr)


class NumWordsMismatchError(Exception):
    
    def __init__(self, word, numMatches):
        super(NumWordsMismatchError, self).__init__()
        self.word = word
        self.numMatches = numMatches
    
    def __str__(self):
        errStr = ("Error: %d matches found in isleDict for '%s'.\n"
                  "Only 1 match allowed--likely you need to break"
                  "up your query text into separate words.")
        return errStr % (self.numMatches, self.word)


class WrongTypeError(Exception):
    
    def __init__(self, errMsg):
        super(WrongTypeError, self).__init__()
        self.str = errMsg
    
    def __str__(self):
        return self.str


class NullPronunciationError(Exception):
    
    def __init__(self, word):
        super(NullPronunciationError, self).__init__()
        self.word = word
    
    def __str__(self):
        return "No pronunciation given for word '%s'" % self.word


class NullPhoneError(Exception):

    def __str(self):
        return "Received an empty phone in the pronunciation list"


def _lcs_lens(xs, ys):
    curr = list(itertools.repeat(0, 1 + len(ys)))
    for x in xs:
        prev = list(curr)
        for i, y in enumerate(ys):
            if x == y:
                curr[i + 1] = prev[i] + 1
            else:
                curr[i + 1] = max(curr[i], prev[i + 1])
    return curr


def _lcs(xs, ys):
    nx, ny = len(xs), len(ys)
    if nx == 0:
        return []
    elif nx == 1:
        return [xs[0]] if xs[0] in ys else []
    else:
        i = nx // 2
        xb, xe = xs[:i], xs[i:]
        ll_b = _lcs_lens(xb, ys)
        ll_e = _lcs_lens(xe[::-1], ys[::-1])
        _, k = max((ll_b[j] + ll_e[ny - j], j)
                   for j in range(ny + 1))
        yb, ye = ys[:k], ys[k:]
        return _lcs(xb, yb) + _lcs(xe, ye)


def _prepPronunciation(phoneList):
    retList = []
    for phone in phoneList:
        
        # Remove diacritics
        for diacritic in isletool.diacriticList:
            phone = phone.replace(diacritic, u'')
        
        # Unify rhotics
        if 'r' in phone:
            phone = 'r'

        phone = phone.lower()

        # Unify vowels
        if isletool.isVowel(phone):
            phone = u'V'
        
        # Only represent the string by its first letter
        try:
            phone = phone[0]
        except IndexError:
            raise NullPhoneError()
        
        # Unify vowels (reducing the vowel to one char)
        if isletool.isVowel(phone):
            phone = u'V'
        
        retList.append(phone)
    
    return retList


def _adjustSyllabification(adjustedPhoneList, syllableList):
    '''
    Inserts spaces into a syllable if needed
    
    Originally the phone list and syllable list contained the same number
    of phones.  But the adjustedPhoneList may have some insertions which are
    not accounted for in the syllableList.
    '''
    i = 0
    retSyllableList = []
    for syllableNum, syllable in enumerate(syllableList):
        j = len(syllable)
        if syllableNum == len(syllableList) - 1:
            j = len(adjustedPhoneList) - i
        tmpPhoneList = adjustedPhoneList[i:i + j]
        numBlanks = -1
        phoneList = tmpPhoneList[:]
        while numBlanks != 0:
            
            numBlanks = tmpPhoneList.count(u"''")
            if numBlanks > 0:
                tmpPhoneList = adjustedPhoneList[i + j:i + j + numBlanks]
                phoneList.extend(tmpPhoneList)
                j += numBlanks
        
        for k, phone in enumerate(phoneList):
            if phone == u"''":
                syllable.insert(k, u"''")
        
        i += j
        
        retSyllableList.append(syllable)
    
    return retSyllableList
        

def _findBestPronunciation(isleWordList, aPron):
    '''
    Words may have multiple candidates in ISLE; returns the 'optimal' one.
    '''
    
    aP = _prepPronunciation(aPron)  # Mapping to simplified phone inventory
    
    numDiffList = []
    withStress = []
    i = 0
    alignedSyllabificationList = []
    alignedActualPronunciationList = []
    for wordTuple in isleWordList:
        aPronMap = copy.deepcopy(aPron)
        syllableList = wordTuple[0]  # syllableList, stressList
        
        iP = [phone for phoneList in syllableList for phone in phoneList]
        iP = _prepPronunciation(iP)

        alignedIP, alignedAP = alignPronunciations(iP, aP)
        
        # Remapping to actual phones
#         alignedAP = [origPronDict.get(phon, u"''") for phon in alignedAP]
        alignedAP = [aPronMap.pop(0) if phon != u"''" else u"''"
                     for phon in alignedAP]
        alignedActualPronunciationList.append(alignedAP)
        
        # Adjusting the syllabification for differences between the dictionary
        # pronunciation and the actual pronunciation
        alignedSyllabification = _adjustSyllabification(alignedIP,
                                                        syllableList)
        alignedSyllabificationList.append(alignedSyllabification)
        
        # Count the number of misalignments between the two
        numDiff = alignedIP.count(u"''") + alignedAP.count(u"''")
        numDiffList.append(numDiff)
        
        # Is there stress in this word
        hasStress = False
        for syllable in syllableList:
            for phone in syllable:
                hasStress = u"ˈ" in phone or hasStress
        
        if hasStress:
            withStress.append(i)
        i += 1
    
    # Return the pronunciation that had the fewest differences
    #     to the actual pronunciation
    minDiff = min(numDiffList)
    
    # When there are multiple candidates that have the minimum number
    #     of differences, prefer one that has stress in it
    bestIndex = None
    bestIsStressed = None
    for i, numDiff in enumerate(numDiffList):
        if numDiff != minDiff:
            continue
        if bestIndex is None:
            bestIndex = i
            bestIsStressed = i in withStress
        else:
            if not bestIsStressed and i in withStress:
                bestIndex = i
                bestIsStressed = True
    
    return (isleWordList, alignedActualPronunciationList,
            alignedSyllabificationList, bestIndex)


def _syllabifyPhones(phoneList, syllableList):
    '''
    Given a phone list and a syllable list, syllabify the phones
    
    Typically used by findBestSyllabification which first aligns the phoneList
    with a dictionary phoneList and then uses the dictionary syllabification
    to syllabify the input phoneList.
    '''
    
    numPhoneList = [len(syllable) for syllable in syllableList]
    
    start = 0
    syllabifiedList = []
    for end in numPhoneList:
        
        syllable = phoneList[start:start + end]
        syllabifiedList.append(syllable)
        
        start += end
    
    return syllabifiedList


def alignPronunciations(phoneListA, phoneListB):
    '''
    Align the phones in two pronunciations

    This will find the longest (non-continuous) common sequence and fill in the gaps
    before, between, and after the characters such that the common elements
    occur at the same points and the character strings are the same length

    In the following example, the phone lists share the sequence ['a', 'd']
    
    ```python
    phoneListA = ['a', 'b', 'c', 'd', 'e', 'f']
    phoneListB = ['l', 'a', 'z', 'd', 'u']
    a, b = alignPronunciations(phoneListA, phoneListB)
    print(a) > ["''", 'a', 'b', 'c', 'd', 'e', 'f']
    print(b) > ['l', 'a', 'z', "''", 'd', 'u', "''"]
    ```
    '''
    
    # First prep the two pronunctions
    phoneListA = [char for char in phoneListA]
    phoneListB = [char for char in phoneListB]
    
    # Remove any elements not in the other list (but maintain order)
    pronATmp = phoneListA
    pronBTmp = phoneListB

    # Find the longest sequence
    sequence = _lcs(pronBTmp, pronATmp)

    # Find the index of the sequence
    # TODO: investigate ambiguous cases
    startA = 0
    startB = 0
    sequenceIndexListA = []
    sequenceIndexListB = []
    for phone in sequence:
        startA = phoneListA.index(phone, startA)
        startB = phoneListB.index(phone, startB)

        sequenceIndexListA.append(startA)
        sequenceIndexListB.append(startB)
    
    # An index on the tail of both will be used to create output strings
    # of the same length
    sequenceIndexListA.append(len(phoneListA))
    sequenceIndexListB.append(len(phoneListB))
    
    # Fill in any blanks such that the sequential items have the same
    # index and the two strings are the same length
    for x in range(len(sequenceIndexListA)):
        indexA = sequenceIndexListA[x]
        indexB = sequenceIndexListB[x]
        if indexA < indexB:
            for x in range(indexB - indexA):
                phoneListA.insert(indexA, "''")
            sequenceIndexListA = [val + indexB - indexA
                                  for val in sequenceIndexListA]
        elif indexA > indexB:
            for x in range(indexA - indexB):
                phoneListB.insert(indexB, "''")
            sequenceIndexListB = [val + indexA - indexB
                                  for val in sequenceIndexListB]
    
    return phoneListA, phoneListB
   

def findBestSyllabification(isleDict, wordText, phoneList):
    '''
    Find the best syllabification for a word

    First find the closest pronunciation to a given pronunciation. Then take
    the syllabification for that pronunciation and map it onto the
    input pronunciation.
    '''
    try:
        phoneList = [unicode(char, "utf-8") for phone in phoneList]
    except (NameError, TypeError):
        pass
    
    isleWordList = isleDict.lookup(wordText)[0]
    
    return _findBestSyllabification(isleWordList, phoneList)
        
    
def _findBestSyllabification(inputIsleWordList, actualPronunciationList):
    '''
    Find the best syllabification for a word
    
    First find the closest pronunciation to a given pronunciation. Then take
    the syllabification for that pronunciation and map it onto the
    input pronunciation.
    '''
    retList = _findBestPronunciation(inputIsleWordList,
                                     actualPronunciationList)
    isleWordList, alignedAPronList, alignedSyllableList, bestIndex = retList
    
    alignedPhoneList = alignedAPronList[bestIndex]
    alignedSyllables = alignedSyllableList[bestIndex]
    syllabification = isleWordList[bestIndex][0]
    stressedSyllableIndexList = isleWordList[bestIndex][1]
    stressedPhoneIndexList = isleWordList[bestIndex][2]
    
    syllableList = _syllabifyPhones(alignedPhoneList, alignedSyllables)
    
    # Get the location of stress in the generated file
    try:
        stressedSyllableI = stressedSyllableIndexList[0]
    except IndexError:
        stressedSyllableI = None
        stressedVowelI = None
    else:
        stressedVowelI = _getSyllableNucleus(syllableList[stressedSyllableI])
    
    # Count the index of the stressed phones, if the stress list has
    # become flattened (no syllable information)
    flattenedStressIndexList = []
    for i, j in zip(stressedSyllableIndexList, stressedPhoneIndexList):
        k = j
        for l in range(i):
            k += len(syllableList[l])
        flattenedStressIndexList.append(k)
    
    return (stressedSyllableI, stressedVowelI, syllableList, syllabification,
            stressedSyllableIndexList, stressedPhoneIndexList,
            flattenedStressIndexList)


def _getSyllableNucleus(phoneList):
    '''
    Given the phones in a syllable, retrieves the vowel index
    '''
    cvList = ['V' if isletool.isVowel(phone) else 'C' for phone in phoneList]
    
    vowelCount = cvList.count('V')
    if vowelCount > 1:
        raise TooManyVowelsInSyllable(phoneList, cvList)
    
    if vowelCount == 1:
        stressI = cvList.index('V')
    else:
        stressI = None
        
    return stressI


def findClosestPronunciation(isleDict, word, phoneList):
    '''
    Find the closest dictionary pronunciation to a provided pronunciation
    '''
    isleWordList = isleDict.lookup(word)


    retList = _findBestPronunciation(isleWordList[0], phoneList)
    isleWordList = retList[0]
    bestIndex = retList[3]
    
    return isleWordList[bestIndex]



#encoding: utf-8
"""
Pysle is an interface to the ISLEX English pronunciation dictionary

ISLEX contains multiple variations in pronunciations for words--like [b ɚ ɹ d] or [b ˈɝ d] for 'bird'.  In contains primary and secondary stress information as in [ˈɛ . l ə . v ˌei . ɾ ɚ]

Using pysle, you can query this resource.

**isletool.py** contains the fundamental parts of the code.  This code is responsible for looking up words and getting their pronunciation. With the search() function, you can look up words based on their pronunication too.  For example, every word that starts with a vowel or that ends in [k æ t] ('scat', 'cat', 'muscat'), etc.

**pronunciationtools.py** contains useful utilities when working with a corpus of pronunciation data.  You can compare pronunciations from your own data with "ideal" pronunciations from the ISLEX dictionary.  Such data can be used to explore speech data--which syllables are expected to be stressed?  Are there phones or syllables that might have been dropped in the pronunciation?

**praattools.py** contains useful utilities if you are working specifically with data stored in TextGrids.  Praat is a speech analysis tool that stores speech annotation data--such as phone labels--in TextGrids.
"""
